/* psoc-ctsic.c
 *
 * Control driver for PSOC Capacitive TouchScreen IC
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/psoc-ctsic.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/jiffies.h>

struct input_dev *vms_input_dev;     /* Representation of an input device */

typedef struct
{
	struct semaphore		mutex;
} psoc_drv_data_t;

static struct _psoc_i2c {
	struct i2c_client	*client;
	struct work_struct	work;
} psoc_i2c;

static struct timer_list mytimer;

static int global_x1, global_y1, global_x2, global_y2;
static char fingers;
static char gesture;

/**
 * psoc_i2c_send - issue a single I2C message in master transmit mode
 * @buf: Data that will be written to the slave
 * @count: How many bytes to write
 *
 * Returns negative errno, or else the number of bytes written.
 */
static int psoc_i2c_send(const char *buf, int count)
{
	int ret = -ENODEV;
	
	psoc_drv_data_t *drv_data = dev_get_drvdata(&psoc_i2c.client->dev);

	BUG_ON(!buf);

	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;
	ret = i2c_master_send(psoc_i2c.client, buf, count);
	up(&drv_data->mutex);

	return ret;
}


/**
 * psoc_i2c_recv - issue a single I2C message in master receive mode
 * @buf: Where to store data read from slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_i2c_recv(char *buf, int count)
{
	int ret = -ENODEV;

	psoc_drv_data_t *drv_data = dev_get_drvdata(&psoc_i2c.client->dev);
	BUG_ON(!buf);

	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;
	ret = i2c_master_recv(psoc_i2c.client, buf, count);
	up(&drv_data->mutex);

	return ret;
}

static void psoc_get_finger_position(void)
{

        char *buff_temp;
        int ret;

        int x1_temp, y1_temp, x2_temp, y2_temp;

        buff_temp = kmalloc(20*sizeof(char), GFP_KERNEL);

        ret = psoc_i2c_recv(buff_temp, 20);

        if (ret != 20){
                printk(KERN_ERR PSOC_DEVNAME ": Error reading finger position, return value of i2c read: [%d] instead of 20\n", ret);
                kfree(buff_temp);
                return;
        }

        //Extract finger position from the i2c packet received from the PSoc
        memcpy(&x1_temp,buff_temp+2,2*sizeof(char));
        memcpy(&y1_temp,buff_temp+4,2*sizeof(char));
        memcpy(&x2_temp,buff_temp+6,2*sizeof(char));
        memcpy(&y2_temp,buff_temp+8,2*sizeof(char));
	memcpy(&fingers,buff_temp+10,sizeof(char));
	memcpy(&gesture,buff_temp+11,sizeof(char));

        // Quantity of fimger detected on the screen
        //memcpy(&current_number_finger, buff_temp+10, sizeof(char));

        //Invert LSB and MSB
        global_x1 = ((x1_temp >> 8) & 0xFF) + ((x1_temp << 8) & 0xFF00);
        global_y1 = ((y1_temp >> 8) & 0xFF) + ((y1_temp << 8) & 0xFF00);
        global_x2 = ((x2_temp >> 8) & 0xFF) + ((x2_temp << 8) & 0xFF00);
        global_y2 = ((y2_temp >> 8) & 0xFF) + ((y2_temp << 8) & 0xFF00);

        //Report position of finger to the input layer
        input_report_abs(vms_input_dev, ABS_Y /*tt Y*/, global_x1 /*psoc X*/);
        input_report_abs(vms_input_dev, ABS_X /*tt X*/, global_y1 /*psoc Y*/);

	input_report_abs(vms_input_dev, ABS_FINGER2_Y /*tt Y*/, global_x2 /*psoc X*/);
	input_report_abs(vms_input_dev, ABS_FINGER2_X /*tt X*/, global_y2 /*psoc Y*/);
	input_report_abs(vms_input_dev, ABS_GESTURE, gesture);
	input_report_abs(vms_input_dev, ABS_FINGERS, fingers);

	if (fingers)
		input_report_abs(vms_input_dev, ABS_PRESSURE, 1);
	else
		input_report_abs(vms_input_dev, ABS_PRESSURE, 0);

        input_sync(vms_input_dev);

        kfree(buff_temp);

}

static void psoc_clear_busy_flag(void)
{
        int ret;
        char buff_clearance[2];

        buff_clearance[0]=1;
        buff_clearance[1]=1;
     
        ret = psoc_i2c_send(buff_clearance, 2);

        if (ret != 2)
                printk(KERN_ERR PSOC_DEVNAME ": Error clearing busy flag, return value of i2c send: [%d] instead of 2\n", ret);
              
}

static void psoc_work(struct work_struct *work)
{
        psoc_get_finger_position();
        psoc_clear_busy_flag();
}


// Timer handler
static void psoc_timer(unsigned long data)
{ 
        schedule_work(&psoc_i2c.work);
        mytimer.expires = jiffies + msecs_to_jiffies(40);
        add_timer(&mytimer);
}

// EINT23 handler (not used for now)
static irqreturn_t psoc_irq_handler(int irq, void *dev_id)
{

        schedule_work(&psoc_i2c.work);
        return IRQ_HANDLED;
};

static void psoc_i2c_init(struct i2c_client *client)
{
        
	BUG_ON(!client);
        
        //Init psoc_i2c struct
	memset(&psoc_i2c, 0x00, sizeof(psoc_i2c));
        psoc_i2c.client = client;
	INIT_WORK(&psoc_i2c.work, psoc_work);

        //Init Timer
        init_timer(&mytimer);
        mytimer.function=&psoc_timer;
        mytimer.expires = jiffies + msecs_to_jiffies(40);

}

static int psoc_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{

	int ret;
	psoc_drv_data_t *drv_data;
	psoc_pdata_t	* pdata;

	printk(KERN_INFO PSOC_DEVNAME " :probing driver\n");
        drv_data = NULL;
        pdata = (psoc_pdata_t *)client->dev.platform_data;

	printk(KERN_INFO PSOC_DEVNAME " :Psoc Capacitive Touch Screen Driver, (C) 2008 TomTom BV\n" );

	if(NULL == pdata) {
		printk(KERN_ERR PSOC_DEVNAME 
				": Missing platform data!\n");
		return -ENODEV;
	}

	BUG_ON(!client);
	psoc_i2c_init(client);

	drv_data = kmalloc(sizeof(psoc_drv_data_t), GFP_KERNEL);

	if(NULL == drv_data) {
		printk(KERN_ERR PSOC_DEVNAME ": Can allocate driver data!\n");
		return -ENOMEM;
	}

	dev_set_drvdata(&client->dev, drv_data);

	init_MUTEX(&drv_data->mutex);

        vms_input_dev = input_allocate_device();

        if (!vms_input_dev) {
                printk(KERN_ERR PSOC_DEVNAME ": Badf init_alloc_device()\n");
                kfree(drv_data);
                return -10;
        }

        printk(KERN_INFO PSOC_DEVNAME ": Input device allocated\n");

        /* Announce that the ts will generate absolute coordinates */
        set_bit(EV_ABS, vms_input_dev->evbit);
        printk(KERN_INFO PSOC_DEVNAME ": Event Abs positions declared\n");

        // Declare max and min of X, Y and pressure
        input_set_abs_params(vms_input_dev, ABS_X, 0, 0x140, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_Y, 0, 0xF0, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_PRESSURE, 0, 1, 0, 0);

	input_set_abs_params(vms_input_dev, ABS_FINGERS, 0, 2, 0, 0);
	input_set_abs_params(vms_input_dev, ABS_FINGER2_X, 0, 0x140, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_FINGER2_Y, 0, 0xF0, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_GESTURE, 0, 0x10, 0, 0);

        vms_input_dev->name = "PSOC Cypress Capacitive TouchScreen";
        vms_input_dev->phys = "input(ts)";
        vms_input_dev->id.vendor = 0xDEAD;
        vms_input_dev->id.product = 0xBEEF;
        vms_input_dev->id.version = 0x0101;

        /* Register with the input subsystem */
        ret = input_register_device(vms_input_dev);

        if (ret) {
                printk(KERN_ERR PSOC_DEVNAME ": Could not register input device(capacitive touchscreen)!\n");
                kfree(drv_data);
                return -EIO;
        }

        printk(KERN_INFO PSOC_DEVNAME ": Input device registered\n");

        // Defined in arch/arm/mach-s3c6410/psoc_ctsic.h (platform device)
        // Declare HPDetect pin as an interrupt (EINT23)
	pdata->reset_ts();

        //Uncomment for use with EINT23

        /*printk(KERN_INFO PSOC_DEVNAME ": irq number [%d]\n", client->irq);
        set_irq_type(client->irq, IRQ_TYPE_EDGE_BOTH);
        if (request_irq(client->irq, &psoc_irq_handler, IRQ_TYPE_EDGE_BOTH, "psoc_1", NULL)) {
                printk(KERN_ERR PSOC_DEVNAME " Could not allocate IRQ (psoc_1)!\n");
                input_unregister_device(vms_input_dev);
                kfree(drv_data);
                return -EIO;
        }*/
 
        add_timer(&mytimer);

	return 0;
}

static int psoc_i2c_remove(struct i2c_client *client)
{
	psoc_drv_data_t *drv_data;

	printk(KERN_INFO PSOC_DEVNAME ": start remove\n");
	drv_data = dev_get_drvdata(&client->dev);
        BUG_ON(!drv_data);

        //Uncomment to use EINT23
        /*disable_irq(client->irq);
        free_irq(client->irq, NULL);*/

        input_unregister_device(vms_input_dev);
	kfree(drv_data);

	return 0;
}

#define psoc_i2c_suspend		NULL
#define psoc_i2c_resume			NULL

static const struct i2c_device_id psoc_id[] = {
	{ PSOC_DEVNAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, psoc_id);

static struct i2c_driver psoc_i2c_driver=
{
	.id		= I2C_DRIVERID_PSOC_CTSIC,
	.probe	= psoc_i2c_probe,
	.remove	= psoc_i2c_remove,
	.suspend= psoc_i2c_suspend,
	.resume	= psoc_i2c_resume,
	.id_table = psoc_id,
	.driver = {
		.name	= PSOC_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static int __init psoc_init(void)
{
	int err;


	printk(KERN_INFO PSOC_DEVNAME " :start init\n");

    if ((err = i2c_add_driver(&psoc_i2c_driver))) {
		printk(KERN_ERR PSOC_DEVNAME ": Could Not Be Added. Err Code: [%i]\n", err);
	}

	return err;
}

static void __exit psoc_exit(void)
{
	printk(KERN_INFO PSOC_DEVNAME " :start exit\n");
	i2c_del_driver(&psoc_i2c_driver);
}


MODULE_AUTHOR("Vincent Dejouy <vincent.dejouy@tomtom.com>");
MODULE_DESCRIPTION("Driver for I2C connected PSoc Capacitive Touchscreen IC.");
MODULE_LICENSE("GPL");

module_init(psoc_init);
module_exit(psoc_exit);
